from .peeling import (data_peeling_algorithm,data_peeling_backward,peeling_to_structure,uniform)
from .plots import (plot_peeling,plot_peeling_nxd,plot_peeling_nxd_back,plot_peeling_nx2,plot_scattermatrix)
from .fuzzy import (samples_to_fuzzy_projection,boxes_to_fuzzy_projection,coverage_samples)
from .examples import (pickle_load,pickle_dump,banana_data,banana_model)